'' SEPLAC (SANNAJ'S EXAMPLE PROGRAMMING LANGUAGE TO ASSEMBLY COMPILER)
'' FREEBASIC_WRITTEN COMPILER PROF OF CONCEPT

''' seplac-lex.bas - char and tolkenwise lexer.

#include once "inc/seplac.bi"               ' Global Header
#include once "inc/seplac-lex.bi"           ' Lexer Header
#include once "inc/seplac-pars.bi"			' Parser Header
#include once "inc/SingleLinkedList.bas"    ' Link List Support

union int_chr
    int as integer
    chr as ubyte
end union

declare function hlexTolken(tolkenN as string, tolkenTclue as integer = toltype_UNKNOWN) as integer

function runLexer(sourcefile as string) as integer
	dim as string localTolken
	dim as byte lexState = lexState_ROOT, NLflag = -1
	dim as int_chr chrPuffer
	dim as integer initSeek, lineSeek = 1
	dim as integer filenum, errornum
	
	filenum = freefile()
	open sourcefile for binary as #filenum
    if err = 2 then
        print_error(,,error_UNEXISTING_INPUT_FILE, " file not found: """ & sourcefile & """")
        close #filenum
        return error_UNEXISTING_INPUT_FILE
    elseif err <> 0 then
        print_error(,,error_OPENING_ERROR, " file couldn't be opened: """ & sourcefile & """")
        close #filenum
        return error_OPENING_ERROR
    end if
    
	do
        if eof(filenum) then
            chrPuffer.chr = 0
        else
            get #filenum,, chrPuffer.chr
        end if
        
        if lexState = lexState_LABEL then			' Inerhalb eines Labels (das nun beendet ist.)
			select case as const chrPuffer.int
			case 0, 32, 9, 13, 10, 35   ' Spezialzeichen EOF, SP, TAB, CR, LF, Kommentarmarker (#)
                localTolken =space(loc(filenum)-initSeek)	' Einlesen Vorbereiten, d.h. Länge des Lables bestimmen.
                get #filenum, initSeek, localTolken ' Einlesen des Labels
                seek filenum, (seek(filenum) + 1)	' Anpassen der Ausleseprosition
                lexState = lexState_ROOT           	' Lexerstatus auf default zurücksetzten.
                errornum = hlexTolken(lcase(localTolken)) ' Call wordlexer
            end select
		end if
		
		if lexState = lexState_DOCU then
			select case as const chrPuffer.int
			case 0, 13, 10   ' Spezialzeichen EOF, CR, LF
				lexState = lexState_ROOT           	' Lexerstatus auf default zurücksetzten.
			end select
		end if
		
		if lexState = lexState_STRING then
			select case as const chrPuffer.int
			case 13, 10		' Spezialzeichen CR, LF
				errornum = hlexTolken(chr(0,0))
			case 34			' Quotation mark
				localTolken =space(seek(filenum)-initSeek)	' Einlesen Vorbereiten, d.h. Länge des Lables bestimmen.
                get #filenum, initSeek, localTolken 		' Einlesen des Labels
                seek filenum, (seek(filenum) + 1)			' Anpassen der Ausleseprosition
                lexState = lexState_ROOT           			' Lexerstatus auf default gesperrt zurücksetzten.
                errornum = hlexTolken(lcase(localTolken), toltype_STRINGLITERAL) 	' Call wordlexer
				continue do									' Skipp Root Lexer for this.
			end select
		end if
		
		if lexState = lexState_Root then
			select case as const chrPuffer.int
			case 0                                 	' Reached end of file.
				errornum = hlexTolken(chr(0))        ' Send eof-Tolken.
				exit do
			case 10, 13                             ' Found new line
				lexState = lexState_ROOT           	' Reset lexer
				NLflag = -1                        	' Set NLflag
				linecount += 1                     	' Increase linecount
				lineSeek = seek(1)                 	' Mark start of new line.
				errornum = hlexTolken(chr(10))       ' Send eol-tolken
			case 35                                	' Init comment (#)
				lexState = lexState_DOCU           	' Select commment mode.
			case 34									' Quotation mark
				initSeek = loc(filenum)         	' Get Prosition of next chr
				lexState = lexState_STRING
			case 32, 9
				rem do nothing
			case else                               ' Sign within a lable
				initSeek = loc(filenum)               ' Get Prosition of this chr
				lexState = lexState_LABEL
			end select
		end if
		
        if errornum <> NO_ERROR then
            close #filenum
            print errornum
            return errornum
        end if
        
    loop
    close #filenum
    return NO_ERROR
end function

function hlexTolken(tolkenN as string, tolkenTclue as integer = toltype_UNKNOWN) as integer
    dim as integer tolkenT, i
    if tolkenTclue = toltype_STRINGLITERAL then
		tolkenT = toltype_STRINGLITERAL
    elseif tolkenN = "print" orelse tolkenN = "stdd" orelse tolkenN = chr(0) orelse tolkenN = chr(10) then
        tolkenT = toltype_KEYWORD
        if tolkenN = chr(0) then tolkenN = "EOF"
        if tolkenN = chr(10) then tolkenN =  "EOL"
    elseif left(tolkenN,1)="""" then
        if right(tolkenN,1)="""" then
            tolkenT = toltype_STRINGLITERAL
        else
            tolkenT = toltype_INVALID
        end if
    elseif instr(left(tolkenN,1), any "0123456789") then        ' Init numeric constant
        tolkenT = toltype_NUMBER
        if len(tolkenN) >= 3 andalso left(tolkenN,1) = "0" then ' Possibly other them decimal system.
            select case tolkenN[1]                              ' Select second caracter as chr
            case asc("x")    ' Hexdecimal
                for i = 2 to len(tolkenN)-1                     ' Check if eveything is a hex digit.
                    if instr(chr(tolkenN[i]), any "0123456789abcdef") then continue for
                    tolkenT = toltype_INVALID
                next i
            case asc("o")    ' Octal
                for i = 2 to len(tolkenN)-1                     ' Check if eveything is a oct digit.
                    if instr(chr(tolkenN[i]), any "012345678") then continue for
                    tolkenT = toltype_INVALID
                next i
            case asc("b")    ' Binary
                for i = 2 to len(tolkenN)-1                     ' Check if eveything is a bin digit.
                    if instr(chr(tolkenN[i]), any "01") then continue for
                    tolkenT = toltype_INVALID
                next i
            case else                                           ' Decimal (Other case)
                goto lexTolken_NUMBER_DECIMAL ' Suspicion faild, so use default treatment.
            end select
            ' Not executed if suspicition failed, case overjumped in this case:
            if tolkenT <> toltype_INVALID then
                tolkenN = str(valint("&" & mid(tolkenN,2)))	' Convert to decimal.
            end if
        else
            lexTolken_NUMBER_DECIMAL:
            for i = 1 to len(tolkenN)-1                         ' Check if eveything is a valid digit.
                if instr(chr(tolkenN[i]), any "0123456789") then continue for
                tolkenT = toltype_INVALID
            next i
        end if
    elseif instr(chr(tolkenN[0]), any "=+-*/") then
		
        if len(tolkenN) > 1 then
            tolkenT = toltype_INVALID
        else
            tolkenT = toltype_OPERATOR
        end if  
    elseif tolkenN = chr(0,0) then
		print_error(inputFile, linecount, error_UNCLOSED_STRINGLITERAL, _
		"Unclosed String Literal found.")
	else
        tolkenT = toltype_VARIABLE
        for i = 0 to len(tolkenN)-1                             ' Check if eveything is a letter.
            if instr(chr(tolkenN[i]), any "0123456789abcdefghijklmnopqrstuvwxyz_") then continue for
            tolkenT = toltype_INVALID
        next i
    end if
    
    if tolkenT = toltype_INVALID then
        print_error(inputFile, linecount, error_INVALID_TOLKEN, _
		"Invalid tolken found: """ & tolkenN & """")
        return error_INVALID_TOLKEN
    else
        return currentParser(trim(tolkenN,""""), tolkenT) ' This function produces the right error codes anyway.
    end if
    
end function
