'' SEPLAC (SANNAJ'S EXAMPLE PROGRAMMING LANGUAGE TO ASSEMBLY COMPILER)
'' FREEBASIC_WRITTEN COMPILER PROF OF CONCEPT

''' seplac-pars.bas - seplac - parser & ast-builder


#define parserRoot_distance 20

#include once "inc/seplac.bi"               ' Global Header
#include once "inc/seplac-pars.bi"          ' Parser
#include once "inc/seplac-symtab.bi"        ' Symbol Table

' Local helper functions
declare function hparserError(tolkenN as string, tolkenT as integer, expected as string) as integer
declare function haddIOControl() as integer
declare function hextractDoubleInput(noTEMP as byte = 0) as integer

' Declare all the parsers:
'declare function parserRoot(tolkenN as string, tolkenT as integer) as integer
declare function parserForStrings(tolkenN as string, tolkenT as integer) as integer
declare function parserForEqualSign(tolkenN as string, tolkenT as integer) as integer
declare function parserForOperant1(tolkenN as string, tolkenT as integer) as integer
declare function parserForOperant2(tolkenN as string, tolkenT as integer) as integer
declare function parserForOperators(tolkenN as string, tolkenT as integer) as integer
declare function parserForEndOfOperation(tolkenN as string, tolkenT as integer) as integer

' Define StringTable for AST
dim shared as byte PrintSTDD = 0
dim shared as zstring ptr STDD = @"stdd"
dim shared currentParser as function(tolkenN as string, tolkenT as integer) as integer = @parserRoot()

function decodeTolkenType(tolkenT as integer) as string
	select case tolkenT
    case toltype_INVALID  : return "invalid tolken"
	case toltype_VARIABLE : return "symbol"
	case toltype_KEYWORD  : return "keyword"
	case toltype_OPERATOR : return "operator"
	case toltype_NUMBER   : return "number literal"
	case toltype_STRINGLITERAL : return "string literal"
	case else : return "undefined type nr. " & tolkenT
	end select
end function

function hextractDoubleInput(noTEMP as byte = 0) as integer
    ' This function is resposible for removing the first input
    dim localInstruction(1 to 2) as bcInstr ptr
    dim tempBuffer as integer = selectedInstruction->operantData(0)
    if noTEMP <> 0 then tempBuffer = getVarID("__seplac_extraTempBuffer")
    if tempBuffer = 0 then return error_INTERNAL_MALLOC
    ' #1 Get local Input
    ' #2 Temporarly store it in result prosition.
    ' #3 Do acctual operation converted into the two operant form.
    ' (#3 is accessed with the selectedInstruction Pointer.)
    ' Setup all the tree isntructions.
    localInstruction(1) = selectedInstruction
    if addBytecodeInstruction() then return error_INTERNAL_MALLOC
    localInstruction(2) = selectedInstruction
    if addBytecodeInstruction() then return error_INTERNAL_MALLOC
    ' First we need to copy the acctuall operation at it's "final" destination.
    *selectedInstruction = *localInstruction(1)
    ' Instruction #1: Get local Input
    with (*localInstruction(1))
        .bcopcode = bcopc_sci0
        ' All the other ones aren't used.
    end with
    ' Instruction #2: Write it to it's desiert destintion
    ' Waring: We need to catch an operation witch uses stdd three times during emitter stage.
    with (*localInstruction(2))
        .bcopcode = bcopc_mov2
        .operantData(0) = tempBuffer
        .operantData(1) = 0
        .operantType(1) = -1
    end with
    ' Instruction #3: Do acctual operation converted into the two operant form.
    with (*selectedInstruction)
        .operantData(1) = tempBuffer
        .operantType(1) = -1
        
        ' No if tempBuffer defaults, operant0 and operant1 are the same, so convert to the 2 operation form.
        if noTEMP = 0 then
            select case .bcopcode
                case bcopc_add3 : .bcopcode = bcopc_add2
                case bcopc_sub3 : .bcopcode = bcopc_sub2
                case bcopc_mul3 : .bcopcode = bcopc_mul2
                case bcopc_div3 : .bcopcode = bcopc_div2
                case else : return error_INTERNAL_VARIOUS
            end select
            .operantData(1) = .operantData(2)
            .operantType(1) = .operantType(2)
            .operantType(2) = -2
        end if
    end with
    return NO_ERROR
end function

function haddIOControl() as integer
    dim doubleInputStatus as byte
    with (*selectedInstruction)
        ' Test whether it's the case, that you will need two inputs the same time.
        if      (.operantData(1) = 0) _
        andalso (.operantData(2) = 0) _
        andalso (.operantType(1) = -1) _
        andalso (.operantType(2) = -1) then
            doubleInputStatus = hextractDoubleInput()
            if doubleInputStatus <> 0 then return doubleInputStatus
        end if
        
        ' This remains at the current instructions prosition.
        dim localInstruction as bcInstr ptr = selectedInstruction
        if ((.operantData(1) = 0) andalso (.operantType(1) = -1)) _
        orelse ((.operantData(2) = 0) andalso (.operantType(2) = -1))then
            ' Requiers an input instruction
            if addBytecodeInstruction() then return error_INTERNAL_MALLOC
            ' The accual operation needs to take place after the input:
            *selectedInstruction = *localInstruction
            ' The add a scan instruction.
            localInstruction->bcopcode = bcopc_sci0
        end if
        if (.operantData(0) = 0) then
            ' Requiers an output instruction
            if addBytecodeInstruction() then return error_INTERNAL_MALLOC
            selectedInstruction->bcopcode = bcopc_pri0
        end if
    end with
end function
    

function parserRoot(tolkenN as string, tolkenT as integer) as integer
    ' This is the global parser.
    select case tolkenT
	case toltype_VARIABLE
        if addBytecodeInstruction() then return error_INTERNAL_MALLOC
        selectedInstruction->operantData(0) = getVarID(tolkenN)
        if selectedInstruction->operantData(0) = 0 then return error_INTERNAL_MALLOC
        currentParser = @parserForEqualSign
	case toltype_KEYWORD
        select case tolkenN
        case "print"
			if addBytecodeInstruction() then return error_INTERNAL_MALLOC
            selectedInstruction->bcopcode = bcopc_prs1
            currentParser = @parserForStrings
		case "EOL", "EOF"
			rem No Operation
		case "stdd"
			PrintSTDD = -1
            if addBytecodeInstruction() then return error_INTERNAL_MALLOC
            selectedInstruction->operantData(0) = 0
			currentParser = @parserForEqualSign
        case else
            return error_INTERNAL_VARIOUS
		end select
		return NO_ERROR
	case else
		return hparserError(tolkenN, tolkenT, "instruction or variable")
	end select
end function

function parserForStrings(tolkenN as string, tolkenT as integer) as integer
    ' This is the lexer for the print Instruction.
    currentParser = @parserRoot
    if toltype_STRINGLITERAL then
        selectedInstruction->operantData(0) = addStringTableEntry(tolkenN)
        if selectedInstruction->operantData(0) = 0 then return error_INTERNAL_MALLOC
        return NO_ERROR
    else
        return hparserError(tolkenN, tolkenT, "instruction or variable")
    end if
end function

function parserForEqualSign(tolkenN as string, tolkenT as integer) as integer
    ' This parser checks for the equal sign withing equasions
    if tolkenN = "=" then
        currentParser = @parserForOperant1
        return NO_ERROR
    else
        return hparserError(tolkenN, tolkenT, "operator ""=""")
    end if
end function

function parserForOperant1(tolkenN as string, tolkenT as integer) as integer
    with (*selectedInstruction)  ' We use the selceted instruction and manipulate it.
        if tolkenT = toltype_NUMBER then
            .operantType(1) = 0  ' Number constant
            .operantData(1) = valint(tolkenN)
        elseif tolkenT = toltype_VARIABLE then
            .operantType(1) = -1
            .operantData(1) = getVarID(tolkenN)
            if .operantData(1) = 0 then return error_INTERNAL_MALLOC
        elseif tolkenN = "stdd" then
            .operantType(1) = -1
            .operantData(1) = 0
        else
            return hparserError(tolkenN, tolkenT, "variable or immediate")
        end if
        if (.operantData(0) = .operantData(1)) andalso (.operantType(1) <> 0) then
            ' Spezial case: Operant 1 = Result:
            .operantType(2) = -2  ' This is indicates that we like to use an two operant format.
            .operantType(1) = 0
            .operantData(1) = 0
        end if
        currentParser = @parserForOperators
        return NO_ERROR
    end with
end function

function parserForOperators(tolkenN as string, tolkenT as integer) as integer
    with (*selectedInstruction)  ' We use the selceted instruction and manipulate it.
        select case tolkenN
        case "+" ' Test whether we need the Composed assign add operators or not
            .bcopcode = iif(.operantType(2) = -2, bcopc_add2, bcopc_add3)
        case "-"
            .bcopcode = iif(.operantType(2) = -2, bcopc_sub2, bcopc_sub3)
        case "*"
            .bcopcode = iif(.operantType(2) = -2, bcopc_mul2, bcopc_mul3)
        case "/"
            .bcopcode = iif(.operantType(2) = -2, bcopc_div2, bcopc_div3)
        case "EOL", "EOF"   ' End of Line meens simple assignment.
            if .operantType(2) = -2 then ' Mov has no combind opcode
                .operantType(1) = -1
                .operantData(1) = .operantData(0)
            end if
            .bcopcode = bcopc_mov2
            currentParser = @parserRoot
            ' We need to add input or output funktions in the case stdd is used.
            return haddIOControl() ' haddIOControl() uses the same error codes, so we can just pass them
        case else
            return hparserError(tolkenN, tolkenT, "operator")
        end select
        currentParser = @parserForOperant2
        return NO_ERROR
    end with
end function

function parserForOperant2(tolkenN as string, tolkenT as integer) as integer
    with (*selectedInstruction)  ' We use the selceted instruction and manipulate it.
        if .operantType(2) = -2 then ' This indicates that we've using a two operant instruction
            if tolkenT = toltype_NUMBER then
                .operantType(1) = 0  ' Number constant
                .operantData(1) = valint(tolkenN)
            elseif tolkenT = toltype_VARIABLE then
                .operantType(1) = -1
                .operantData(1) = getVarID(tolkenN)
                if .operantData(1) = 0 then return error_INTERNAL_MALLOC
            elseif tolkenN = "stdd" then
                .operantType(1) = -1
                .operantData(1) = 0
            else
                goto Syntax_error_occured
            end if
        else
            if tolkenT = toltype_NUMBER then
                .operantType(2) = 0  ' Number constant
                .operantData(2) = valint(tolkenN)
            elseif tolkenT = toltype_VARIABLE then
                .operantType(2) = -1
                .operantData(2) = getVarID(tolkenN)
                if .operantData(2) = 0 then return error_INTERNAL_MALLOC
            elseif tolkenN = "stdd" then
                .operantType(2) = -1
                .operantData(2) = 0
            else
                goto Syntax_error_occured
            end if
        end if
        currentParser = @parserForEndOfOperation
        ' We need to add input or output funktions in the case stdd is used.
        return haddIOControl() ' haddIOControl() uses the same error codes, so we can just pass them
        
        Syntax_error_occured:
        return hparserError(tolkenN, tolkenT, "variable or immediate")
    end with
end function

function parserForEndOfOperation(tolkenN as string, tolkenT as integer) as integer
    select case tolkenN
    case "EOL", "EOF"
        currentParser = @parserRoot
        return NO_ERROR
    case else
        return hparserError(tolkenN, tolkenT, "EOL")
    end select
end function

function hparserError(tolkenN as string, tolkenT as integer, expected as string) as integer
    ' This returns all the errors. Do not use error codes!
    select case tolkenT
    case toltype_INVALID
		print_error(inputFile, linecount, error_INVALID_TOLKEN, _
		"Invalid Tolken """ & tolkenN & """")
		return error_INVALID_TOLKEN
	case toltype_VARIABLE
        print_error(inputFile, linecount, error_UNEXPECTET_VARIABLE, _
		"Expect " & expected & " found variable: """ & tolkenN & """")
		return error_UNEXPECTET_VARIABLE
	case toltype_KEYWORD
        print_error(inputFile, linecount, error_UNEXPECTET_KEYWORD, _
		"Expect " & expected & " found keyword: """ & tolkenN & """")
		return error_UNEXPECTET_KEYWORD
	case toltype_OPERATOR
		print_error(inputFile, linecount, error_UNEXPECTET_OP, _
		"Expect " & expected & " found operator: """ & tolkenN & """")
		return error_UNEXPECTET_OP
	case toltype_NUMBER
		print_error(inputFile, linecount, error_UNEXPECTET_NUM, _
		"Expect " & expected & " found number: """ & tolkenN & """")
		return error_UNEXPECTET_NUM
	case toltype_STRINGLITERAL
		print_error(inputFile, linecount, error_UNEXPECTET_STRING, _
		"Expect " & expected & " found string literal: """ & tolkenN & """")
		return error_UNEXPECTET_STRING
	case else
		print "Internal Error: Undefined Tolkentype";
		print "tolkenT = #" & tolkenT
		print "tolkenN = """ & tolkenN & """"
		return error_INTERNAL_VARIOUS
	end select
end function
