'' SEPLAC (SANNAJ'S EXAMPLE PROGRAMMING LANGUAGE TO ASSEMBLY COMPILER)
'' FREEBASIC_WRITTEN COMPILER PROF OF CONCEPT
''
'' PROGRAMMING LANGUAGE:
''
'' Syntax           	Function
'' var1 = var2 + var3	' Add var2 and var3 and store result in var1
'' var1 = var2 - var3   ' Sub var3 from var3 and stroe result in var1
'' var1 = var2 * var3   ' Multiply var2 with var3 and store result in var1
'' var1 = var2 / var3   ' Divide var2 by var3 and store result in var1
'' var1 = var2			' Copy value from var2 to var1
'' stdd = var2      	' Output var2
'' var1 = stdd      	' Input number to var1
'' print "text1"    	' Print text1.
'' #<Info>          	' A comment
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

'' Compilation options:
'' -c 		Compile and assemble only, do not link at all.
'' -r		Compile only, do not assemble.
'' -h		Print some helpfull informations.
'' -v		Be verbious
'' -version Print version

''' seplac.bas - Main programm part:

' TODO: SPLIT this into different files:

#include once "inc/seplac.bi"               ' Global Header
#include once "inc/seplac-lex.bi"			' Lexer & Parser
#include once "inc/seplac-emitter.bi"       ' Assembly Code Generator

dim shared as compileControlFlagType compileControlFlag
dim shared as string inputFile
dim shared as integer linecount = 1

sub print_error(filename as string = "", linenum as uinteger = 0, errornum as integer, description as string)
    if filename = "" then
        print "error " & errornum & ": " & description
    elseif linenum = 0 then
        print filename & " error " & errornum & ": " & description
    else
        print filename & "(" & linenum & ") error " & errornum & ": " & description
    end if
    'end errornum
end sub

sub hcmd_lexer_printHelp()
	print "Usage: seplac [options] inputfile"
	print ""
	print "options:"
	print "-c 		    Compile and assemble only, do not link at all."
	print "-h		    Print some helpfull informations and exit."
	print "-r		    Compile only, do not assemble."
	print "-v		    Be verbose"
	print "--version    Show compiler version"
	print "--noclean	Do not remove any tempory outputs"
end sub

function seplac_main() as integer
	dim as integer i = 1
    dim as integer localErr
    dim as string anyFile, cmdParams       ' Filename of the product without any extention.
    
    '---------------------------------------------------------------------------
    ' Step 1: Interpret command line
    
	do
        select case command(i)
		case ""	' No further options, exit lexer
			exit do
		case "-h"
			hcmd_lexer_printHelp()
            return NO_ERROR
		case "-version"
			print "Version " & SEPLAC_VERSION
            return NO_ERROR
		case "-r"
			compileControlFlag.outType = outType_Asm
		case "-c"
			compileControlFlag.outType = outType_Obj
		case "-v"
			compileControlFlag.verbose = true
		case "--noclean"
			compileControlFlag.noClean = true
		case else
			if inputFile = "" then
				inputFile = command(i)
                inputFile = trim(inputFile, """")   ' Remove string markers.
			else
				print_error(,,error_INVALID_CMD_ARG, "Invalid command-line option, """ & command(i) & """ found.")
                return error_INVALID_CMD_ARG
			end if
		end select
        i += 1
	loop
	if inputFile = "" then
        hcmd_lexer_printHelp()
        return 1
    end if
    
    
    '---------------------------------------------------------------------------
    ' Step 2: Compile into assembly
    
    anyFile = left(inputFile, instrrev(inputFile, ".")-1)   ' Remove file extetion
    if anyFile = "" then anyFile = inputFile                ' Catch no ending error.
    
    printIfFlag(verbose) !"compiling:\t" & inputFile & " -o " & anyFile & ".asm"
    
    runLexer(inputFile)				' lexical scan and parse input file and set up AST
    
    localErr = emitToFile(anyFile & ".asm")                 ' Emit file
    if localErr <> NO_ERROR then return localErr
    
    '---------------------------------------------------------------------------
    ' Step 3: Assemble
    
    ' If Assembly code is wished, you need to leave now
    if compileControlFlag.outType = outType_Asm then return NO_ERROR
    
    cmdParams = " --32 --strip-local-absolute " & anyFile & ".asm -o " & anyFile & ".o"
    printIfFlag(verbose) !"assembling:\t" & AS_EXE & cmdParams
    if exec(AS_EXE, cmdParams) then return error_HELPER_PROGRAM_ABORT
    if compileControlFlag.noClean = false then kill anyFile & ".asm"
    
    '---------------------------------------------------------------------------
    ' Step 4: Link into executable
    
    ' If an Object File is wished, you need to leave now
    if compileControlFlag.outType = outType_Obj then return NO_ERROR
    
    cmdParams = " -dynamic-linker /lib/ld-linux.so.2 -lc " & anyFile & ".o " & RuntimeLibName & " -o " & anyFile & EXECUTABLE_ENDING
    printIfFlag(verbose) !"linking:\t" & LD_EXE & cmdParams 
    
    if exec(LD_EXE, cmdParams) then return error_HELPER_PROGRAM_ABORT
    if compileControlFlag.noClean = false then kill anyFile & ".o"
    
    return NO_ERROR
    
end function

end seplac_main()

